require 'rails/generators/base'

%w( levels level responses association routes attributes attribute navigation ).each do |file|
  require File.join(File.dirname(__FILE__), file)
end



module Schofield

  module Generators

    class Schofield < Rails::Generators::Base

      source_root File.expand_path(File.join(File.dirname(__FILE__), '../', 'templates'))

      class_option :tables_to_ignore, :aliases => '-i', :type => :string, :default => '', :desc => 'Indicates which tables to ignore'
      class_option :ask_questions,    :aliases => '-a', :type => :boolean, :desc => 'Indicates whether to ask questions again rather than use recorded answers'



      def generate
        Responses.generator    = self
        Responses.re_ask       = options[:ask_questions]
        Level.tables_to_ignore = tables_to_ignore
        Levels.models          = models
        Responses.save
        Association.reference_parents
        Responses.save
        Association.report
      end

      def generate_routes
        inject_into_file(File.join(Rails.root, 'config', 'routes.rb'), Routes.generate, :before => /end\s*\Z/)
      end

      def generate_controllers
        Levels.all.select(&:controllers?).each do |level|
          source = level.join? ? 'join_controller.erb' : 'controller.erb'
          destination = File.join('app/controllers/admin', "#{level.name.tableize}_controller.rb")
          template_with_context(source, Proc.new{level}, destination)
        end
      end

      def generate_views
        Levels.all.select(&:views?).each do |level|
          template_with_context('form.erb', Proc.new{level}, view_path(level, '_form.haml'))
          create_file(view_path(level, 'new.haml'), %q(= render :partial => 'form', :layout => 'admin/shared/module_layout'))
          template_with_context('show.erb', Proc.new{level}, view_path(level, 'show.haml'))
          if !level.nested?
            template_with_context('index.erb', Proc.new{level}, view_path(level, 'index.haml'))
          end
        end
      end

      def generate_models
      Levels.all.select(&:models?).each do |level|
          destination = File.join('app/models', "#{level.name}.rb")
          template_with_context('model.erb', Proc.new{level}, destination)
        end
      end

      def generate_navigation
        inject_into_file(File.join(Rails.root, 'app/views/admin/shared', '_navigation.haml'), Navigation.generate, :after => /- if current_user\n/)
      end

      def generate_tables
        tables = []
        Levels.all.select(&:tables?).each do |level|
          columns_level = level.subclass? ? level.superclass : level
          columns = columns_level.attributes.select{ |a| columns_level.form_field?(a) && !a.text? }.sort_by(&:weight).map(&:to_column)
          if level.nested?
            tables << <<-STRING.gsub(/^ {12}/, '')
              def #{level.name.pluralize}_table
                [
                  #{columns.join(",\n                  ")}
                ]
              end
            STRING
          else
            tables << <<-STRING.gsub(/^ {12}/, '')
              def render_#{level.name.pluralize} collection
                table_renderer(:#{level.name.pluralize}, collection).render(
                  #{columns.join(",\n                  ")}
                )
              end
            STRING
          end
        end
        create_file 'app/helpers/admin/table_definitions_helper.rb' do
          "module Admin::TableDefinitionsHelper\n\n#{tables.join("\n\n")}\nend"
        end
      end



      private

      def tables_to_ignore
        @tables_to_ignore ||= %w( schema_migrations ) + options[:tables_to_ignore].split(',').map{ |e| e.strip.tableize }
      end

      def view_path level, file_name
        File.join('app/views/admin', level.name.tableize, file_name)
      end

      def models
        @models ||= tables.map { |t| t.camelize.singularize.constantize }
      end

      def tables
        ActiveRecord::Base.connection.tables - tables_to_ignore
      end

      def template_with_context(source, context, destination=nil, config={}, &block)
        destination ||= source
        source  = File.expand_path(find_in_source_paths(source.to_s))
        create_file destination, nil, config do
          content = ERB.new(::File.binread(source), nil, '-').result(context)
          content = block.call(content) if block
          content
        end
      end

    end
  end
end